PROJECT_ENV_NAME = "NEPTUNE_PROJECT"

API_TOKEN_ENV_NAME = "NEPTUNE_API_TOKEN"

DISABLE_COLORS = "NEPTUNE_DISABLE_COLORS"

DEBUG_MODE = "NEPTUNE_DEBUG_MODE"

SUBPROCESS_KILL_TIMEOUT = "NEPTUNE_SUBPROCESS_KILL_TIMEOUT"

ALLOW_SELF_SIGNED_CERTIFICATE = "NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE"



__all__ = ("logger",)

import logging
import os

from neptune_scale.core.styles import (
    STYLES,
    ensure_style_detected,
)
from neptune_scale.envs import DEBUG_MODE

NEPTUNE_LOGGER_NAME = "neptune"
NEPTUNE_DEBUG_FILE_NAME = "neptune.log"
LOG_FORMAT = "{blue}%(name)s{end} :: {bold}%(levelname)s{end} :: %(message)s"
DEBUG_FORMAT = "%(asctime)s :: %(name)s :: %(levelname)s :: %(processName)s(%(process)d):%(threadName)s:%(filename)s:%(funcName)s():%(lineno)d %(message)s"


def get_logger() -> logging.Logger:
    ensure_style_detected()

    neptune_logger = logging.getLogger(NEPTUNE_LOGGER_NAME)
    neptune_logger.setLevel(logging.INFO)

    if os.environ.get(DEBUG_MODE, "False").lower() in ("true", "1"):
        neptune_logger.setLevel(logging.DEBUG)

        file_handler = logging.FileHandler(NEPTUNE_DEBUG_FILE_NAME)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(logging.Formatter(DEBUG_FORMAT))
        neptune_logger.addHandler(file_handler)

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(logging.Formatter(LOG_FORMAT.format(**STYLES)))
    neptune_logger.addHandler(stream_handler)

    return neptune_logger


logger = get_logger()



from __future__ import annotations

__all__ = ("OperationsQueue",)

from multiprocessing import Queue
from time import monotonic
from typing import (
    TYPE_CHECKING,
    Optional,
)

from neptune_scale.core.components.abstract import Resource
from neptune_scale.core.components.queue_element import SingleOperation
from neptune_scale.core.logger import logger
from neptune_scale.core.validation import verify_type
from neptune_scale.parameters import (
    MAX_MULTIPROCESSING_QUEUE_SIZE,
    MAX_QUEUE_ELEMENT_SIZE,
    MAX_QUEUE_SIZE,
)

if TYPE_CHECKING:
    from threading import RLock

    from neptune_api.proto.neptune_pb.ingest.v1.pub.ingest_pb2 import RunOperation


class OperationsQueue(Resource):
    def __init__(
        self,
        *,
        lock: RLock,
        max_size: int = MAX_QUEUE_SIZE,
    ) -> None:
        verify_type("max_size", max_size, int)

        self._lock: RLock = lock
        self._max_size: int = max_size

        self._sequence_id: int = 0
        self._last_timestamp: Optional[float] = None
        self._queue: Queue[SingleOperation] = Queue(maxsize=min(MAX_MULTIPROCESSING_QUEUE_SIZE, max_size))

    @property
    def queue(self) -> Queue[SingleOperation]:
        return self._queue

    @property
    def last_sequence_id(self) -> int:
        with self._lock:
            return self._sequence_id - 1

    @property
    def last_timestamp(self) -> Optional[float]:
        with self._lock:
            return self._last_timestamp

    def enqueue(self, *, operation: RunOperation, size: Optional[int] = None, key: Optional[float] = None) -> None:
        try:
            is_metadata_update = operation.HasField("update")
            serialized_operation = operation.SerializeToString()

            if len(serialized_operation) > MAX_QUEUE_ELEMENT_SIZE:
                raise ValueError(f"Operation size exceeds the maximum allowed size ({MAX_QUEUE_ELEMENT_SIZE})")

            with self._lock:
                self._last_timestamp = monotonic()
                # TODO: should we not block here, and just call the error callback if we were to block?
                self._queue.put(
                    SingleOperation(
                        sequence_id=self._sequence_id,
                        timestamp=self._last_timestamp,
                        operation=serialized_operation,
                        metadata_size=size,
                        is_batchable=is_metadata_update,
                        operation_key=key,
                    ),
                    block=True,
                    timeout=None,
                )
                self._sequence_id += 1
        except Exception as e:
            logger.error("Failed to enqueue operation: %s %s", e, operation)
            raise e

    def close(self) -> None:
        self._queue.close()
        # This is needed to avoid hanging the main process
        self._queue.cancel_join_thread()



from __future__ import annotations

__all__ = ("SyncProcess",)

import multiprocessing
import queue
import signal
import threading
from multiprocessing import (
    Process,
    Queue,
)
from multiprocessing.sharedctypes import Synchronized
from multiprocessing.synchronize import Condition
from types import FrameType
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Literal,
    NamedTuple,
    Optional,
    Type,
    TypeVar,
)

import backoff
import httpx
from neptune_api.errors import (
    InvalidApiTokenException,
    UnableToDeserializeApiKeyError,
    UnableToExchangeApiKeyError,
    UnableToRefreshTokenError,
    UnexpectedStatus,
)
from neptune_api.proto.google_rpc.code_pb2 import Code
from neptune_api.proto.neptune_pb.ingest.v1.ingest_pb2 import IngestCode
from neptune_api.proto.neptune_pb.ingest.v1.pub.client_pb2 import (
    BulkRequestStatus,
    RequestId,
)
from neptune_api.proto.neptune_pb.ingest.v1.pub.ingest_pb2 import RunOperation

from neptune_scale.api.api_client import (
    ApiClient,
    backend_factory,
)
from neptune_scale.core.components.abstract import (
    Resource,
    WithResources,
)
from neptune_scale.core.components.aggregating_queue import AggregatingQueue
from neptune_scale.core.components.daemon import Daemon
from neptune_scale.core.components.errors_tracking import ErrorsQueue
from neptune_scale.core.components.queue_element import (
    BatchedOperations,
    SingleOperation,
)
from neptune_scale.core.logger import logger
from neptune_scale.core.util import safe_signal_name
from neptune_scale.exceptions import (
    NeptuneConnectionLostError,
    NeptuneFieldPathEmpty,
    NeptuneFieldPathExceedsSizeLimit,
    NeptuneFieldPathInvalid,
    NeptuneFieldPathNonWritable,
    NeptuneFieldTypeConflicting,
    NeptuneFieldTypeUnsupported,
    NeptuneFloatValueNanInfUnsupported,
    NeptuneInternalServerError,
    NeptuneInvalidCredentialsError,
    NeptuneOperationsQueueMaxSizeExceeded,
    NeptuneProjectInvalidName,
    NeptuneProjectNotFound,
    NeptuneRetryableError,
    NeptuneRunConflicting,
    NeptuneRunDuplicate,
    NeptuneRunForkParentNotFound,
    NeptuneRunInvalidCreationParameters,
    NeptuneRunNotFound,
    NeptuneSeriesPointDuplicate,
    NeptuneSeriesStepNonIncreasing,
    NeptuneSeriesStepNotAfterForkPoint,
    NeptuneSeriesTimestampDecreasing,
    NeptuneStringSetExceedsSizeLimit,
    NeptuneStringValueExceedsSizeLimit,
    NeptuneSynchronizationStopped,
    NeptuneUnableToAuthenticateError,
    NeptuneUnauthorizedError,
    NeptuneUnexpectedError,
    NeptuneUnexpectedResponseError,
)
from neptune_scale.parameters import (
    INTERNAL_QUEUE_FEEDER_THREAD_SLEEP_TIME,
    MAX_QUEUE_SIZE,
    MAX_REQUEST_RETRY_SECONDS,
    MAX_REQUESTS_STATUS_BATCH_SIZE,
    SHUTDOWN_TIMEOUT,
    STATUS_TRACKING_THREAD_SLEEP_TIME,
    SYNC_PROCESS_SLEEP_TIME,
    SYNC_THREAD_SLEEP_TIME,
)

T = TypeVar("T")


CODE_TO_ERROR: Dict[IngestCode.ValueType, Optional[Type[Exception]]] = {
    IngestCode.OK: None,
    IngestCode.PROJECT_NOT_FOUND: NeptuneProjectNotFound,
    IngestCode.PROJECT_INVALID_NAME: NeptuneProjectInvalidName,
    IngestCode.RUN_NOT_FOUND: NeptuneRunNotFound,
    IngestCode.RUN_DUPLICATE: NeptuneRunDuplicate,
    IngestCode.RUN_CONFLICTING: NeptuneRunConflicting,
    IngestCode.RUN_FORK_PARENT_NOT_FOUND: NeptuneRunForkParentNotFound,
    IngestCode.RUN_INVALID_CREATION_PARAMETERS: NeptuneRunInvalidCreationParameters,
    IngestCode.FIELD_PATH_EXCEEDS_SIZE_LIMIT: NeptuneFieldPathExceedsSizeLimit,
    IngestCode.FIELD_PATH_EMPTY: NeptuneFieldPathEmpty,
    IngestCode.FIELD_PATH_INVALID: NeptuneFieldPathInvalid,
    IngestCode.FIELD_PATH_NON_WRITABLE: NeptuneFieldPathNonWritable,
    IngestCode.FIELD_TYPE_UNSUPPORTED: NeptuneFieldTypeUnsupported,
    IngestCode.FIELD_TYPE_CONFLICTING: NeptuneFieldTypeConflicting,
    IngestCode.SERIES_POINT_DUPLICATE: NeptuneSeriesPointDuplicate,
    IngestCode.SERIES_STEP_NON_INCREASING: NeptuneSeriesStepNonIncreasing,
    IngestCode.SERIES_STEP_NOT_AFTER_FORK_POINT: NeptuneSeriesStepNotAfterForkPoint,
    IngestCode.SERIES_TIMESTAMP_DECREASING: NeptuneSeriesTimestampDecreasing,
    IngestCode.FLOAT_VALUE_NAN_INF_UNSUPPORTED: NeptuneFloatValueNanInfUnsupported,
    IngestCode.STRING_VALUE_EXCEEDS_SIZE_LIMIT: NeptuneStringValueExceedsSizeLimit,
    IngestCode.STRING_SET_EXCEEDS_SIZE_LIMIT: NeptuneStringSetExceedsSizeLimit,
}


class StatusTrackingElement(NamedTuple):
    sequence_id: int
    timestamp: float
    request_id: str


def code_to_exception(code: IngestCode.ValueType) -> Optional[Type[Exception]]:
    if code in CODE_TO_ERROR:
        return CODE_TO_ERROR[code]
    return NeptuneUnexpectedError


class PeekableQueue(Generic[T]):
    def __init__(self) -> None:
        self._lock: threading.RLock = threading.RLock()
        self._queue: queue.Queue[T] = queue.Queue()

    def put(self, element: T) -> None:
        with self._lock:
            self._queue.put(element)

    def peek(self, max_size: int) -> Optional[List[T]]:
        with self._lock:
            size = self._queue.qsize()
            if size == 0:
                return None

            items = []
            for i in range(min(size, max_size)):
                item = self._queue.queue[i]
                items.append(item)
            return items

    def commit(self, n: int) -> None:
        with self._lock:
            size = self._queue.qsize()
            for _ in range(min(size, n)):
                self._queue.get()


def with_api_errors_handling(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except (InvalidApiTokenException, UnableToDeserializeApiKeyError):
            raise NeptuneInvalidCredentialsError()
        except (UnableToRefreshTokenError, UnableToExchangeApiKeyError, UnexpectedStatus):
            raise NeptuneUnableToAuthenticateError()
        except (httpx.ConnectError, httpx.TimeoutException, httpx.RemoteProtocolError):
            raise NeptuneConnectionLostError()
        except Exception as e:
            raise e

    return wrapper


class SyncProcess(Process):
    def __init__(
        self,
        operations_queue: Queue,
        errors_queue: ErrorsQueue,
        api_token: str,
        project: str,
        family: str,
        mode: Literal["async", "disabled"],
        last_put_seq: Synchronized[int],
        last_put_seq_wait: Condition,
        last_ack_seq: Synchronized[int],
        last_ack_seq_wait: Condition,
        last_ack_timestamp: Synchronized[float],
        last_ack_timestamp_wait: Condition,
        max_queue_size: int = MAX_QUEUE_SIZE,
    ) -> None:
        super().__init__(name="SyncProcess")

        self._external_operations_queue: Queue[SingleOperation] = operations_queue
        self._errors_queue: ErrorsQueue = errors_queue
        self._api_token: str = api_token
        self._project: str = project
        self._family: str = family
        self._last_put_seq: Synchronized[int] = last_put_seq
        self._last_put_seq_wait: Condition = last_put_seq_wait
        self._last_ack_seq: Synchronized[int] = last_ack_seq
        self._last_ack_seq_wait: Condition = last_ack_seq_wait
        self._last_ack_timestamp: Synchronized[float] = last_ack_timestamp
        self._last_ack_timestamp_wait: Condition = last_ack_timestamp_wait
        self._max_queue_size: int = max_queue_size
        self._mode: Literal["async", "disabled"] = mode

        # This flag is set when a termination signal is caught
        self._stop_event = multiprocessing.Event()

    def _handle_signal(self, signum: int, frame: Optional[FrameType]) -> None:
        logger.debug(f"Received signal {safe_signal_name(signum)}")
        self._stop_event.set()  # Trigger the stop event

    def run(self) -> None:
        logger.info("Data synchronization started")

        # Register signals handlers
        signal.signal(signal.SIGTERM, self._handle_signal)

        worker = SyncProcessWorker(
            project=self._project,
            family=self._family,
            api_token=self._api_token,
            errors_queue=self._errors_queue,
            external_operations_queue=self._external_operations_queue,
            last_put_seq=self._last_put_seq,
            last_put_seq_wait=self._last_put_seq_wait,
            last_ack_seq=self._last_ack_seq,
            last_ack_seq_wait=self._last_ack_seq_wait,
            max_queue_size=self._max_queue_size,
            last_ack_timestamp=self._last_ack_timestamp,
            last_ack_timestamp_wait=self._last_ack_timestamp_wait,
            mode=self._mode,
        )
        worker.start()
        try:
            while not self._stop_event.is_set():
                worker.join(timeout=SYNC_PROCESS_SLEEP_TIME)
        except KeyboardInterrupt:
            logger.debug("Data synchronization interrupted by user")
        finally:
            logger.info("Data synchronization stopping")
            worker.interrupt()
            worker.wake_up()
            worker.join(timeout=SHUTDOWN_TIMEOUT)
            worker.close()
        logger.info("Data synchronization finished")


class SyncProcessWorker(WithResources):
    def __init__(
        self,
        *,
        api_token: str,
        project: str,
        family: str,
        mode: Literal["async", "disabled"],
        errors_queue: ErrorsQueue,
        external_operations_queue: multiprocessing.Queue[SingleOperation],
        last_put_seq: Synchronized[int],
        last_put_seq_wait: Condition,
        last_ack_seq: Synchronized[int],
        last_ack_seq_wait: Condition,
        last_ack_timestamp: Synchronized[float],
        last_ack_timestamp_wait: Condition,
        max_queue_size: int = MAX_QUEUE_SIZE,
    ) -> None:
        self._errors_queue = errors_queue

        self._internal_operations_queue: AggregatingQueue = AggregatingQueue(max_queue_size=max_queue_size)
        self._status_tracking_queue: PeekableQueue[StatusTrackingElement] = PeekableQueue()
        self._sync_thread = SenderThread(
            api_token=api_token,
            operations_queue=self._internal_operations_queue,
            status_tracking_queue=self._status_tracking_queue,
            errors_queue=self._errors_queue,
            family=family,
            last_put_seq=last_put_seq,
            last_put_seq_wait=last_put_seq_wait,
            mode=mode,
        )
        self._external_to_internal_thread = InternalQueueFeederThread(
            external=external_operations_queue,
            internal=self._internal_operations_queue,
            errors_queue=self._errors_queue,
        )
        self._status_tracking_thread = StatusTrackingThread(
            api_token=api_token,
            mode=mode,
            project=project,
            errors_queue=self._errors_queue,
            status_tracking_queue=self._status_tracking_queue,
            last_ack_seq=last_ack_seq,
            last_ack_seq_wait=last_ack_seq_wait,
            last_ack_timestamp=last_ack_timestamp,
            last_ack_timestamp_wait=last_ack_timestamp_wait,
        )

    @property
    def threads(self) -> tuple[Daemon, ...]:
        return self._external_to_internal_thread, self._sync_thread, self._status_tracking_thread

    @property
    def resources(self) -> tuple[Resource, ...]:
        return self._external_to_internal_thread, self._sync_thread, self._status_tracking_thread

    def interrupt(self) -> None:
        for thread in self.threads:
            thread.interrupt()

    def wake_up(self) -> None:
        for thread in self.threads:
            thread.wake_up()

    def start(self) -> None:
        for thread in self.threads:
            thread.start()

    def join(self, timeout: Optional[int] = None) -> None:
        # The same timeout will be applied to each thread separately
        for thread in self.threads:
            thread.join(timeout=timeout)


class InternalQueueFeederThread(Daemon, Resource):
    def __init__(
        self,
        external: multiprocessing.Queue[SingleOperation],
        internal: AggregatingQueue,
        errors_queue: ErrorsQueue,
    ) -> None:
        super().__init__(name="InternalQueueFeederThread", sleep_time=INTERNAL_QUEUE_FEEDER_THREAD_SLEEP_TIME)

        self._external: multiprocessing.Queue[SingleOperation] = external
        self._internal: AggregatingQueue = internal
        self._errors_queue: ErrorsQueue = errors_queue

        self._latest_unprocessed: Optional[SingleOperation] = None

    def get_next(self) -> Optional[SingleOperation]:
        if self._latest_unprocessed is not None:
            return self._latest_unprocessed

        try:
            self._latest_unprocessed = self._external.get(timeout=INTERNAL_QUEUE_FEEDER_THREAD_SLEEP_TIME)
            return self._latest_unprocessed
        except queue.Empty:
            return None

    def commit(self) -> None:
        self._latest_unprocessed = None

    def work(self) -> None:
        try:
            while not self._is_interrupted():
                operation = self.get_next()
                if operation is None:
                    continue

                try:
                    self._internal.put_nowait(operation)
                    self.commit()
                except queue.Full:
                    logger.debug("Internal queue is full (%d elements), waiting for free space", self._internal.maxsize)
                    self._errors_queue.put(NeptuneOperationsQueueMaxSizeExceeded(max_size=self._internal.maxsize))
                    # Sleep before retry
                    break
        except Exception as e:
            self._errors_queue.put(e)
            self.interrupt()
            raise NeptuneSynchronizationStopped() from e


class SenderThread(Daemon, WithResources):
    def __init__(
        self,
        api_token: str,
        family: str,
        operations_queue: AggregatingQueue,
        status_tracking_queue: PeekableQueue[StatusTrackingElement],
        errors_queue: ErrorsQueue,
        last_put_seq: Synchronized[int],
        last_put_seq_wait: Condition,
        mode: Literal["async", "disabled"],
    ) -> None:
        super().__init__(name="SenderThread", sleep_time=SYNC_THREAD_SLEEP_TIME)

        self._api_token: str = api_token
        self._family: str = family
        self._operations_queue: AggregatingQueue = operations_queue
        self._status_tracking_queue: PeekableQueue[StatusTrackingElement] = status_tracking_queue
        self._errors_queue: ErrorsQueue = errors_queue
        self._last_put_seq: Synchronized[int] = last_put_seq
        self._last_put_seq_wait: Condition = last_put_seq_wait
        self._mode: Literal["async", "disabled"] = mode

        self._backend: Optional[ApiClient] = None
        self._latest_unprocessed: Optional[BatchedOperations] = None

    def get_next(self) -> Optional[BatchedOperations]:
        if self._latest_unprocessed is not None:
            return self._latest_unprocessed

        try:
            self._latest_unprocessed = self._operations_queue.get()
            return self._latest_unprocessed
        except queue.Empty:
            return None

    def commit(self) -> None:
        self._latest_unprocessed = None

    @property
    def resources(self) -> tuple[Resource, ...]:
        if self._backend is not None:
            return (self._backend,)
        return ()

    @backoff.on_exception(backoff.expo, NeptuneConnectionLostError, max_time=MAX_REQUEST_RETRY_SECONDS)
    @with_api_errors_handling
    def submit(self, *, operation: RunOperation) -> Optional[RequestId]:
        if self._backend is None:
            self._backend = backend_factory(api_token=self._api_token, mode=self._mode)

        response = self._backend.submit(operation=operation, family=self._family)

        if response.status_code == 403:
            raise NeptuneUnauthorizedError()

        if response.status_code != 200:
            logger.error("HTTP response error: %s", response.status_code)
            if response.status_code // 100 == 5:
                raise NeptuneInternalServerError()
            else:
                raise NeptuneUnexpectedResponseError()

        return response.parsed

    def work(self) -> None:
        try:
            while (operation := self.get_next()) is not None:
                sequence_id, timestamp, data = operation

                try:
                    logger.debug("Submitting operation #%d with size of %d bytes", sequence_id, len(data))
                    run_operation = RunOperation()
                    run_operation.ParseFromString(data)
                    request_id = self.submit(operation=run_operation)

                    if request_id is None:
                        raise NeptuneUnexpectedError("Server response is empty")

                    logger.debug("Operation #%d submitted as %s", sequence_id, request_id.value)
                    self._status_tracking_queue.put(
                        StatusTrackingElement(sequence_id=sequence_id, request_id=request_id.value, timestamp=timestamp)
                    )
                    self.commit()
                except NeptuneRetryableError as e:
                    self._errors_queue.put(e)
                    # Sleep before retry
                    break

                # Update Last PUT sequence id and notify threads in the main process
                with self._last_put_seq_wait:
                    self._last_put_seq.value = sequence_id
                    self._last_put_seq_wait.notify_all()
        except Exception as e:
            self._errors_queue.put(e)
            self._last_put_seq_wait.notify_all()
            self.interrupt()
            raise NeptuneSynchronizationStopped() from e


class StatusTrackingThread(Daemon, WithResources):
    def __init__(
        self,
        api_token: str,
        mode: Literal["async", "disabled"],
        project: str,
        errors_queue: ErrorsQueue,
        status_tracking_queue: PeekableQueue[StatusTrackingElement],
        last_ack_seq: Synchronized[int],
        last_ack_seq_wait: Condition,
        last_ack_timestamp: Synchronized[float],
        last_ack_timestamp_wait: Condition,
    ) -> None:
        super().__init__(name="StatusTrackingThread", sleep_time=STATUS_TRACKING_THREAD_SLEEP_TIME)

        self._api_token: str = api_token
        self._mode: Literal["async", "disabled"] = mode
        self._project: str = project
        self._errors_queue: ErrorsQueue = errors_queue
        self._status_tracking_queue: PeekableQueue[StatusTrackingElement] = status_tracking_queue
        self._last_ack_seq: Synchronized[int] = last_ack_seq
        self._last_ack_seq_wait: Condition = last_ack_seq_wait
        self._last_ack_timestamp: Synchronized[float] = last_ack_timestamp
        self._last_ack_timestamp_wait: Condition = last_ack_timestamp_wait

        self._backend: Optional[ApiClient] = None

    @property
    def resources(self) -> tuple[Resource, ...]:
        if self._backend is not None:
            return (self._backend,)
        return ()

    def get_next(self) -> Optional[List[StatusTrackingElement]]:
        try:
            return self._status_tracking_queue.peek(max_size=MAX_REQUESTS_STATUS_BATCH_SIZE)
        except queue.Empty:
            return None

    @backoff.on_exception(backoff.expo, NeptuneConnectionLostError, max_time=MAX_REQUEST_RETRY_SECONDS)
    @with_api_errors_handling
    def check_batch(self, *, request_ids: List[str]) -> Optional[BulkRequestStatus]:
        if self._backend is None:
            self._backend = backend_factory(api_token=self._api_token, mode=self._mode)

        response = self._backend.check_batch(request_ids=request_ids, project=self._project)

        if response.status_code == 403:
            raise NeptuneUnauthorizedError()

        if response.status_code != 200:
            logger.error("HTTP response error: %s", response.status_code)
            if response.status_code // 100 == 5:
                raise NeptuneInternalServerError()
            else:
                raise NeptuneUnexpectedResponseError()

        return response.parsed

    def work(self) -> None:
        try:
            while (batch := self.get_next()) is not None:
                request_ids = [element.request_id for element in batch]
                sequence_ids = [element.sequence_id for element in batch]
                timestamps = [element.timestamp for element in batch]

                try:
                    response = self.check_batch(request_ids=request_ids)
                    if response is None:
                        raise NeptuneUnexpectedError("Server response is empty")
                except NeptuneRetryableError as e:
                    self._errors_queue.put(e)
                    # Small give up, sleep before retry
                    break

                operations_to_commit, processed_sequence_id, processed_timestamp = 0, None, None
                for request_status, request_sequence_id, timestamp in zip(response.statuses, sequence_ids, timestamps):
                    if any(code_status.code == Code.UNAVAILABLE for code_status in request_status.code_by_count):
                        logger.debug(f"Operation #{request_sequence_id} is not yet processed.")
                        # Request status not ready yet, sleep and retry
                        break

                    for code_status in request_status.code_by_count:
                        if code_status.code != Code.OK and (error := code_to_exception(code_status.detail)) is not None:
                            self._errors_queue.put(error())

                    operations_to_commit += 1
                    processed_sequence_id, processed_timestamp = request_sequence_id, timestamp

                if operations_to_commit > 0:
                    self._status_tracking_queue.commit(operations_to_commit)

                    # Update Last ACK sequence id and notify threads in the main process
                    if processed_sequence_id is not None:
                        logger.debug(f"Operations up to #{processed_sequence_id} are completed.")

                        with self._last_ack_seq_wait:
                            self._last_ack_seq.value = processed_sequence_id
                            self._last_ack_seq_wait.notify_all()

                    # Update Last ACK timestamp and notify threads in the main process
                    if processed_timestamp is not None:
                        with self._last_ack_timestamp_wait:
                            self._last_ack_timestamp.value = processed_timestamp
                            self._last_ack_timestamp_wait.notify_all()
                else:
                    # Sleep before retry
                    break
        except Exception as e:
            self._errors_queue.put(e)
            self.interrupt()
            self._last_ack_seq_wait.notify_all()
            raise NeptuneSynchronizationStopped() from e



__all__ = ["Daemon"]

import abc
import threading
from enum import Enum

from neptune_scale.core.logger import logger


class Daemon(threading.Thread):
    class DaemonState(Enum):
        INIT = 1
        WORKING = 2
        PAUSING = 3
        PAUSED = 4
        INTERRUPTED = 5
        STOPPED = 6

    def __init__(self, sleep_time: float, name: str) -> None:
        super().__init__(daemon=True, name=name)
        self._sleep_time = sleep_time
        self._state: Daemon.DaemonState = Daemon.DaemonState.INIT
        self._wait_condition = threading.Condition()

    def interrupt(self) -> None:
        logger.debug(f"Thread {self} interrupted.")
        with self._wait_condition:
            self._state = Daemon.DaemonState.INTERRUPTED
            self._wait_condition.notify_all()

    def pause(self) -> None:
        with self._wait_condition:
            if self._state != Daemon.DaemonState.PAUSED:
                if not self._is_interrupted():
                    self._state = Daemon.DaemonState.PAUSING
                self._wait_condition.notify_all()
                self._wait_condition.wait_for(lambda: self._state != Daemon.DaemonState.PAUSING)

    def resume(self) -> None:
        with self._wait_condition:
            if not self._is_interrupted():
                self._state = Daemon.DaemonState.WORKING
            self._wait_condition.notify_all()

    def wake_up(self) -> None:
        with self._wait_condition:
            self._wait_condition.notify_all()

    def disable_sleep(self) -> None:
        self._sleep_time = 0

    def is_running(self) -> bool:
        with self._wait_condition:
            return self._state in (
                Daemon.DaemonState.WORKING,
                Daemon.DaemonState.PAUSING,
                Daemon.DaemonState.PAUSED,
            )

    def _is_interrupted(self) -> bool:
        with self._wait_condition:
            return self._state in (Daemon.DaemonState.INTERRUPTED, Daemon.DaemonState.STOPPED)

    def run(self) -> None:
        with self._wait_condition:
            if not self._is_interrupted():
                self._state = Daemon.DaemonState.WORKING
        try:
            while not self._is_interrupted():
                with self._wait_condition:
                    if self._state == Daemon.DaemonState.PAUSING:
                        self._state = Daemon.DaemonState.PAUSED
                        self._wait_condition.notify_all()
                        self._wait_condition.wait_for(lambda: self._state != Daemon.DaemonState.PAUSED)

                if self._state == Daemon.DaemonState.WORKING:
                    self.work()
                    with self._wait_condition:
                        if self._sleep_time > 0 and self._state == Daemon.DaemonState.WORKING:
                            self._wait_condition.wait(timeout=self._sleep_time)
        finally:
            with self._wait_condition:
                self._state = Daemon.DaemonState.STOPPED
                self._wait_condition.notify_all()

            logger.debug(f"Thread {self} is finished.")

    @abc.abstractmethod
    def work(self) -> None: ...



from __future__ import annotations

__all__ = (
    "NeptuneScaleError",
    "NeptuneScaleWarning",
    "NeptuneOperationsQueueMaxSizeExceeded",
    "NeptuneUnauthorizedError",
    "NeptuneInvalidCredentialsError",
    "NeptuneUnexpectedError",
    "NeptuneConnectionLostError",
    "NeptuneUnableToAuthenticateError",
    "NeptuneRetryableError",
    "NeptuneUnexpectedResponseError",
    "NeptuneInternalServerError",
    "NeptuneProjectNotFound",
    "NeptuneProjectInvalidName",
    "NeptuneRunNotFound",
    "NeptuneRunDuplicate",
    "NeptuneRunConflicting",
    "NeptuneRunForkParentNotFound",
    "NeptuneRunInvalidCreationParameters",
    "NeptuneFieldPathExceedsSizeLimit",
    "NeptuneFieldPathEmpty",
    "NeptuneFieldPathInvalid",
    "NeptuneFieldPathNonWritable",
    "NeptuneFieldTypeUnsupported",
    "NeptuneFieldTypeConflicting",
    "NeptuneSeriesPointDuplicate",
    "NeptuneSeriesStepNonIncreasing",
    "NeptuneSeriesStepNotAfterForkPoint",
    "NeptuneSeriesTimestampDecreasing",
    "NeptuneFloatValueNanInfUnsupported",
    "NeptuneStringValueExceedsSizeLimit",
    "NeptuneStringSetExceedsSizeLimit",
    "NeptuneSynchronizationStopped",
    "NeptuneAsyncLagThresholdExceeded",
    "NeptuneProjectNotProvided",
    "NeptuneApiTokenNotProvided",
)

from typing import Any

from neptune_scale.core.styles import (
    STYLES,
    ensure_style_detected,
)


class NeptuneScaleError(Exception):
    message = "An error occurred in the Neptune Scale client."

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        ensure_style_detected()
        super().__init__(self.message.format(*args, **STYLES, **kwargs))


class NeptuneScaleWarning(Warning):
    message = "A warning occurred in the Neptune Scale client."

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        ensure_style_detected()
        super().__init__(self.message.format(*args, **STYLES, **kwargs))


class NeptuneSynchronizationStopped(NeptuneScaleError):
    message = "Internal synchronization process was stopped."


class NeptuneOperationsQueueMaxSizeExceeded(NeptuneScaleError):
    message = """
{h1}
----NeptuneOperationsQueueMaxSizeExceeded--------------------------------------
{end}
The queue size for internal operations was exceeded because too much data was queued in a short time.

The synchronization is paused until the queue size drops below the maximum.

To resolve this issue, consider the following:
    - Reduce the frequency of data being sent to the queue, or throttle the rate of operations.
    - Cautiously increase the queue size through the `max_queue_size` argument.
        Note: To ensure that memory usage remains within acceptable limits,
        closely monitor your system's memory consumption.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneUnauthorizedError(NeptuneScaleError):
    message = """
{h1}
----NeptuneUnauthorizedError---------------------------------------------------
{end}
You don't have permission to access the given resource.

    - Verify that your API token is correct. To find your API token:
        - Log in to Neptune Scale and open the user menu.
        - If your workspace uses service accounts, ask the project owner to provide the token.

    - Verify that the provided project name is correct.
      The correct project name should look like this: {correct}WORKSPACE_NAME/PROJECT_NAME{end}
      It has two parts:
          - {correct}WORKSPACE_NAME{end}: can be your username or your organization name
          - {correct}PROJECT_NAME{end}: the name specified for the project

   - Ask your workspace admin to grant you the necessary privileges to the project.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneInvalidCredentialsError(NeptuneScaleError):
    message = """
{h1}
----NeptuneInvalidCredentialsError---------------------------------------------
{end}
The provided API token is invalid.
Make sure you copied your API token while logged in to Neptune Scale.
If your workspace uses service accounts, ask the project owner for the token.

There are two options to provide the API token:
    - Set it as an environment variable in your operating system
    - Paste it into your Python code (not recommended)

{h2}Environment variable{end} {correct}(Recommended){end}
Set the NEPTUNE_API_TOKEN environment variable depending on your operating system:

    {correct}Linux/Unix{end}
    In the terminal:
        {bash}export NEPTUNE_API_TOKEN="YOUR_API_TOKEN"{end}

    {correct}Windows{end}
    In Command Prompt or similar:
        {bash}setx NEPTUNE_API_TOKEN "YOUR_API_TOKEN"{end}

and omit the {bold}api_token{end} argument from the {bold}Run{end} constructor:
    {python}neptune_scale.Run(project="WORKSPACE_NAME/PROJECT_NAME"){end}

{h2}Option 2: Run argument{end}
Pass the token to the {bold}Run{end} constructor via the {bold}api_token{end} argument:
    {python}neptune_scale.Run(project="WORKSPACE_NAME/PROJECT_NAME", api_token="YOUR_API_TOKEN"){end}

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneUnexpectedError(NeptuneScaleError):
    message = """
{h1}
----NeptuneUnexpectedError-----------------------------------------------------
{end}
An unexpected error occurred in the Neptune Scale client. For help, contact support@neptune.ai. Raw exception name: "{reason}"

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""

    def __init__(self, reason: str) -> None:
        super().__init__(reason=reason)


class NeptuneRetryableError(NeptuneScaleError):
    pass


class NeptuneConnectionLostError(NeptuneRetryableError):
    message = """
{h1}
----NeptuneConnectionLostError-------------------------------------------------
{end}
The connection to the Neptune server was lost. Ensure that your computer is connected to the internet and that
    firewall settings aren't blocking the connection.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneUnexpectedResponseError(NeptuneRetryableError):
    message = """
{h1}
----NeptuneUnexpectedResponseError-------------------------------------------------
{end}
The Neptune server returned an unexpected response. This is a temporary problem.
If the problem persists, please contact us.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneInternalServerError(NeptuneRetryableError):
    message = """
{h1}
----NeptuneInternalServerError-------------------------------------------------
{end}
We have encountered an internal server error. If the problem persists, please contact us.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneUnableToAuthenticateError(NeptuneScaleError):
    message = """
{h1}
----NeptuneUnableToAuthenticateError-------------------------------------------
{end}
The client was unable to authenticate with the Neptune server. Ensure that your API token is correct.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneProjectNotFound(NeptuneScaleError):
    message = """
{h1}
----NeptuneProjectNotFound-----------------------------------------------------
{end}
Project not found. Either the project hasn't been created yet or the name is incorrect.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneProjectInvalidName(NeptuneScaleError):
    message = """
{h1}
----NeptuneProjectInvalidName--------------------------------------------------
{end}
Project name is either empty or too long.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneRunNotFound(NeptuneScaleError):
    message = """
{h1}
----NeptuneRunNotFound---------------------------------------------------------
{end}
Run not found. May happen when the run is not yet created.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneRunDuplicate(NeptuneScaleWarning):
    message = """
{h1}
----NeptuneRunDuplicate--------------------------------------------------------
{end}
Identical run already exists. If you wanted to resume an existing run, include the argument `resume=True`.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneRunConflicting(NeptuneScaleError):
    message = """
{h1}
----NeptuneRunConflicting------------------------------------------------------
{end}
Run with specified `run_id` already exists, but has different creation parameters (`family` or `fork_run_id`).

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneRunForkParentNotFound(NeptuneScaleWarning):
    message = """
{h1}
----NeptuneRunForkParentNotFound-----------------------------------------------
{end}
Missing fork parent run.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneRunInvalidCreationParameters(NeptuneScaleError):
    message = """
{h1}
----NeptuneRunInvalidCreationParameters----------------------------------------
{end}
Invalid run creation parameters. For example, the experiment name is too large.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneFieldPathExceedsSizeLimit(NeptuneScaleError):
    message = """
{h1}
----NeptuneFieldPathExceedsSizeLimit-------------------------------------------
{end}
Field path is too long. Maximum length is 1024 bytes (not characters).

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneFieldPathEmpty(NeptuneScaleError):
    message = """
{h1}
----NeptuneFieldPathEmpty------------------------------------------------------
{end}
Field path is empty.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneFieldPathInvalid(NeptuneScaleError):
    message = """
{h1}
----NeptuneFieldPathInvalid----------------------------------------------------
{end}
Field path is invalid. To troubleshoot the problem, ensure that the UTF-8 encoding is valid.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneFieldPathNonWritable(NeptuneScaleError):
    message = """
{h1}
----NeptuneFieldPathNonWritable------------------------------------------------
{end}
Field path is non-writable. Some special sys/ fields are read-only.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneFieldTypeUnsupported(NeptuneScaleError):
    message = """
{h1}
----NeptuneFieldTypeUnsupported------------------------------------------------
{end}
Field type is not supported by the system.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneFieldTypeConflicting(NeptuneScaleError):
    message = """
{h1}
----NeptuneFieldTypeConflicting------------------------------------------------
{end}
Field type is different from the one that was previously logged for this series.
Once a field type is set, it cannot be changed.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneSeriesPointDuplicate(NeptuneScaleWarning):
    message = """
{h1}
----NeptuneSeriesPointDuplicate------------------------------------------------
{end}
The exact same data point was already logged for this series.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneSeriesStepNonIncreasing(NeptuneScaleError):
    message = """
{h1}
----NeptuneSeriesStepNonIncreasing---------------------------------------------
{end}
The step of a series value is smaller than the most recently logged step for this series or the step is exactly the same,
  but the value is different.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneSeriesStepNotAfterForkPoint(NeptuneScaleError):
    message = """
{h1}
----NeptuneSeriesStepNotAfterForkPoint-----------------------------------------
{end}
The series value must be greater than the step specified by the `fork_step` argument.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneSeriesTimestampDecreasing(NeptuneScaleError):
    message = """
{h1}
----NeptuneSeriesTimestampDecreasing-------------------------------------------
{end}
The timestamp of a series value is less than the most recently logged value. Identical timestamps are allowed.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneFloatValueNanInfUnsupported(NeptuneScaleError):
    message = """
{h1}
----NeptuneFloatValueNanInfUnsupported-----------------------------------------
{end}
Unsupported value type for float64 field or float64 series. Applies to Inf and NaN values.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneStringValueExceedsSizeLimit(NeptuneScaleError):
    message = """
{h1}
----NeptuneStringValueExceedsSizeLimit-----------------------------------------
{end}
String value is too long. Maximum length is 64KB.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneStringSetExceedsSizeLimit(NeptuneScaleError):
    message = """
{h1}
----NeptuneStringSetExceedsSizeLimit-------------------------------------------
{end}
String Set value is too long. Maximum length is 64KB.

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneAsyncLagThresholdExceeded(NeptuneScaleError):
    message = """
    {h1}
    ----NeptuneAsyncLagThresholdExceeded----------------------------------------
    {end}
    Neptune is experiencing a high delay in synchronizing data.

    {correct}Need help?{end}-> Contact support@neptune.ai

    Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
    """


class NeptuneProjectNotProvided(NeptuneRetryableError):
    message = """
{h1}
----NeptuneProjectNotProvided--------------------------------------------------
{end}
The project name was not provided. Make sure to specify the project name in the `project` parameter of the `Run`
constructor or with the `NEPTUNE_PROJECT` environment variable.

For instructions, see https://docs-beta.neptune.ai/setup

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""


class NeptuneApiTokenNotProvided(NeptuneRetryableError):
    message = """
{h1}
----NeptuneApiTokenNotProvided-------------------------------------------------
{end}
The Neptune API token was not provided. Make sure to specify the API token in the `api_token` parameter of the `Run`
constructor or with the `NEPTUNE_API_TOKEN` environment variable.

For instructions, see https://docs-beta.neptune.ai/api_token

{correct}Need help?{end}-> Contact support@neptune.ai

Struggling with the formatting? To disable it, set the `NEPTUNE_DISABLE_COLORS` environment variable to `True`.
"""



from __future__ import annotations

__all__ = (
    "make_value",
    "make_step",
    "datetime_to_proto",
    "pb_key_size",
)

from datetime import datetime
from typing import (
    List,
    Set,
    Union,
)

from google.protobuf.timestamp_pb2 import Timestamp
from neptune_api.proto.neptune_pb.ingest.v1.common_pb2 import (
    Step,
    StringSet,
    Value,
)


def make_value(value: Union[Value, float, str, int, bool, datetime, List[str], Set[str]]) -> Value:
    if isinstance(value, Value):
        return value
    if isinstance(value, float):
        return Value(float64=value)
    elif isinstance(value, bool):
        return Value(bool=value)
    elif isinstance(value, int):
        return Value(int64=value)
    elif isinstance(value, str):
        return Value(string=value)
    elif isinstance(value, datetime):
        return Value(timestamp=datetime_to_proto(value))
    elif isinstance(value, (list, set)):
        fv = Value(string_set=StringSet(values=value))
        return fv
    else:
        raise ValueError(f"Unsupported ingest field value type: {type(value)}")


def datetime_to_proto(dt: datetime) -> Timestamp:
    dt_ts = dt.timestamp()
    return Timestamp(seconds=int(dt_ts), nanos=int((dt_ts % 1) * 1e9))


def make_step(number: Union[float, int], raise_on_step_precision_loss: bool = False) -> Step:
    """
    Converts a number to protobuf Step value. Example:
    >>> assert make_step(7.654321, True) == Step(whole=7, micro=654321)
    Args:
        number: step expressed as number
        raise_on_step_precision_loss: inform converter whether it should silently drop precision and
            round down to 6 decimal places or raise an error.

    Returns: Step protobuf used in Neptune API.
    """
    m = int(1e6)
    micro: int = int(number * m)
    if raise_on_step_precision_loss and number * m - micro != 0:
        raise ValueError(f"step must not use more than 6-decimal points, got: {number}")

    whole = micro // m
    micro = micro % m

    return Step(whole=whole, micro=micro)


def pb_key_size(key: str) -> int:
    """
    Calculates the size of the string in the protobuf message including an overhead of the length prefix (varint)
        with an assumption of maximal string length.
    """
    key_bin = bytes(key, "utf-8")
    return len(key_bin) + 2 + (1 if len(key_bin) > 127 else 0)
